unit PatternEngine;

{ Pattern String Engine 1.3
  for Borland Delphi
  Freeware with source
  Copyright (c) 1997-2003, SoftComplete Development
  Web:   http://www.textolution.com
         http://www.softcomplete.com
  Email: support@textolution.com

  THIS SOFTWARE AND THE ACCOMPANYING FILES ARE DISTRIBUTED 
  "AS IS" AND WITHOUT WARRANTIES AS TO PERFORMANCE OF MERCHANTABILITY OR 
  ANY OTHER WARRANTIES WHETHER EXPRESSED OR IMPLIED.
  NO WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE IS OFFERED.
  THE USER MUST ASSUME THE ENTIRE RISK OF USING THE ACCOMPANYING CODE.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions: 

  1. The origin of this software must not be misrepresented, you must 
     not claim that you wrote the original software. If you use this software 
     in a product, an acknowledgment in the product documentation 
     would be appreciated but is not required. 
  2. Altered source versions must be plainly marked as such, and must not be 
     misrepresented as being the original software.
  3. Original copyright may not be removed or altered from any source 
     distribution.
}

interface

uses  SysUtils, Windows, Classes;

type
  PatternException = class(Exception);
  TPattern = class;

  TPatItemType = (piNil,piSimple,piPattern);
  
  TPatItem = class
  private
    function GetMatchLen: integer;
  public
    ItemType: TPatItemType;
    pSimple:   string;
    pPattern:  TPattern;
    constructor Create(const VarRec: TVarRec);
    destructor Destroy; override;
    property MatchLen: integer read GetMatchLen;
  end;

  TPatList = class(TList)
  private
    Owner: TPattern;
    function GetItems(Index: integer): TPatItem;
  public
    constructor Create(AOwner: TPattern; const Pattern: array of const);
    destructor Destroy; override;
    procedure NewItem(const VarRec: TVarRec);
    property Items[Index: integer]: TPatItem read GetItems; default;
  end;

  TPattern = class(TObject)
  private
    Line: string;
    function GetSetValue(var AValue: string): TPattern;
    function GetMatchStr: string;
    function GetMatchNextPtr: PChar;
  protected
    PStack: integer;
    Pattern: TPatList;
    Value: PString;
    OldValue: string;
    PStart,PS: PChar;
    CanResume: Boolean;
    FMatch: Boolean;
    FMatchLen: integer;
    procedure MatchOK;
    procedure MatchFail;
    procedure BackValue;
    procedure Init; virtual;
    function CompareStr(S: PChar; Len: integer): Boolean; virtual;
  public
    constructor Create(const APattern: array of const);
    destructor Destroy; override;

    procedure Activate(const S: string);
    procedure ActivatePtr(P: PChar);

    function Match: Boolean;
    procedure MatchNext; virtual; abstract;

    property MatchPtr: PChar read PS;
    property MatchNextPtr: PChar read GetMatchNextPtr;
    property MatchStr: string read GetMatchStr;
    property MatchLen: integer read FMatchLen;
    property IsMatch: Boolean read FMatch;

    property SetValue[var AValue: string]: TPattern read GetSetValue; default;
  end;

  TPSE = class(TPattern)
  public
    procedure MatchNext; override;
  end;

  TPSEOneOf = class(TPattern)
  public
    procedure MatchNext; override;
  end;

  TPSEMayBe = class(TPSE)
  private
    SecondResume: Boolean;
  protected
    procedure Init; override;
  public
    procedure MatchNext; override;
  end;

  TPSECharSet = class(TPattern)
  private
    CharSet: set of Char;
    MaxLen: integer;
  public
    // SCharSet:: 'asd..g0..5.8' -> ['a','s','d'..'g','0'..'5','.','8']
    // if #0 in CharSet then goal may be empty length
    constructor Create(const SCharSet: string);
    constructor CreateLen(const SCharSet: string; AMaxLen: integer);
    procedure MatchNext; override;
  end;

  TPSEVar = class(TPattern)
  private
    Value: PString;
  public
    constructor Create(var AValue: string);
    procedure MatchNext; override;
  end;

  TPSESetLen = class(TPSE)
  private
    MinLen,MaxLen: integer;
  public
    constructor Create(const APattern: array of const; Min,Max: integer);
    procedure MatchNext; override;
  end;

  TPatternCallBack = function: Boolean;

  TPSEBack = class(TPattern)
  private
    CallBack: TPatternCallBack;
  public
    constructor Create(ACallBack: TPatternCallBack);
    procedure MatchNext; override;
  end;

// sequential group of patterns
function PSE(const APattern: array of const): TPattern;

// APattern[1] or APattern[2] or APattern[3] or ...
function PSEOneOf(const APattern: array of const): TPattern;

// APattern or nothink
function PSEMayBe(const APattern: array of const): TPattern;

// any string with symbols from charset
function PSECharSet(const SCharSet: string): TPattern;

// any string with symbols from charset and with length less or eq to AMaxLen
function PSECharSetLen(const SCharSet: string; AMaxLen: integer): TPattern;

// single char from charset
function PSEChar(const SCharSet: string): TPattern;

// string eq to AValue.
function PSEVar(var AValue: string): TPattern;

// Pattern, but with chars count in range Min..Max
function PSESetLen(const APattern: array of const; Min,Max: integer): TPattern;

// Pattern with user defined callback
function PSEBack(ACallBack: TPatternCallBack): TPattern;

// some frequently used predefined patterns:
function PSEDigit: TPattern;    // 0..9
function PSEDigits: TPattern;   // <digit>{<digit>}
function PSEInteger: TPattern;  // [+-]<digits>
function PSENumber: TPattern;   // [+-]<digits>[.<digits>]
function PSEIdent: TPattern;    // a..zA..Z_{a..zA..Z0..9_}
function PSEAny: TPattern;      // any string or empty
function PSEAnyStr: TPattern;   // any string with length > 0
function PSEAnyChar: TPattern;  // any char

implementation

function PSEBack(ACallBack: TPatternCallBack): TPattern;
begin
  Result:=TPSEBack.Create(ACallBack);
end;

function PSEAnyStr: TPattern;
begin
  Result:=PSECharSet(#1'..'#255);
end;

function PSEAnyChar: TPattern;
begin
  Result:=PSECharSetLen(#1'..'#255,1);
end;

function PSEAny: TPattern;
begin
  Result:=PSECharSet(#1'..'#255#0);
end;

function PSESetLen(const APattern: array of const; Min,Max: integer): TPattern;
begin
  Result:=TPSESetLen.Create(APattern,Min,Max);
end;

function PSEMayBe(const APattern: array of const): TPattern;
begin
  Result:=TPSEMayBe.Create(APattern);
end;

function PSEVar(var AValue: string): TPattern;
begin
  Result:=TPSEVar.Create(AValue);
end;

function PSENumber: TPattern;
begin
  Result:=PSE([PSEChar('+-'#0),PSEDigits,PSEMayBe([PSEMayBe([DecimalSeparator]),PSEDigits])]);
end;

function PSEInteger: TPattern;
begin
  Result:=PSE([PSEChar('+-'#0),PSEDigits]);
end;

function PSEIdent: TPattern;
begin
  Result:=PSE([PSEChar('a..zA..Z_'),PSECharSet('a..zA..Z_0..9'#0)]);
end;

function PSEDigits: TPattern;
begin
  Result:=PSECharSet('0..9');
end;

function PSEDigit: TPattern;
begin
  Result:=PSEChar('0..9');
end;

function PSE(const APattern: array of const): TPattern;
begin
  Result:=TPSE.Create(APattern);
end;

function PSEOneOf(const APattern: array of const): TPattern;
begin
  Result:=TPSEOneOf.Create(APattern);
end;

function PSEChar(const SCharSet: string): TPattern;
begin
  Result:=TPSECharSet.CreateLen(SCharSet,1);
end;

function PSECharSet(const SCharSet: string): TPattern;
begin
  Result:=TPSECharSet.Create(SCharSet);
end;

function PSECharSetLen(const SCharSet: string; AMaxLen: integer): TPattern;
begin
  Result:=TPSECharSet.CreateLen(SCharSet,AMaxLen);
end;

//------------------------------ TPSEBack --------------------------------------

constructor TPSEBack.Create(ACallBack: TPatternCallBack);
begin
  inherited Create([nil]);
  CallBack:=ACallBack;
end;

procedure TPSEBack.MatchNext;
begin
  if Assigned(CallBack) then
    if CallBack then MatchOK
    else MatchFail
  else MatchOK;
  CanResume:=False;
end;

//---------------------------- TPSESetLen --------------------------------------

constructor TPSESetLen.Create(const APattern: array of const; Min,Max: integer);
var  t: integer;
begin
  inherited Create(APattern);
  if Min < 0 then Min:=-1;
  if Max < 0 then Max:=-1;
  if (Min >= 0) and (Max >= 0) and (Min > Max) then begin
    t:=Min; Min:=Max; Max:=t;
  end;
  MinLen:=Min; MaxLen:=Max;
end;

procedure TPSESetLen.MatchNext;
begin
  repeat
    inherited MatchNext;
    if not FMatch then Break;
    if MinLen >= 0 then
      FMatch:=FMatch and (MatchLen >= MinLen);
    if MaxLen >= 0 then
      FMatch:=FMatch and (MatchLen <= MaxLen);
  until FMatch;
end;

//------------------------------- TPSEVar --------------------------------------

constructor TPSEVar.Create(var AValue: string);
begin
  inherited Create([nil]);
  Value:=@AValue;
end;

procedure TPSEVar.MatchNext;
begin
  if CompareStr(PChar(Value^),Length(Value^)) then begin
    PS:=PS+Length(Value^);
    MatchOK;
  end else
    MatchFail;
  CanResume:=False;
end;

//------------------------------- TPSEMayBe ------------------------------------

procedure TPSEMayBe.MatchNext;
begin
  if not CanResume then MatchOK
  else begin
    CanResume:=SecondResume;
    inherited MatchNext;
    SecondResume:=True;
  end;
end;

procedure TPSEMayBe.Init;
begin
  inherited Init;
  SecondResume:=False;
end;

//------------------------------ TPSECharSet -----------------------------------

procedure TPSECharSet.MatchNext;
begin
  if (PS-PStart) >= MaxLen then
    MatchFail
  else
    if not CanResume and (#0 in CharSet) then
      MatchOK
    else if (PS^ <> #0) and (PS^ in CharSet) then begin
      PS:=PS+1;
      MatchOK;
      CanResume:=((PS-PStart) < MaxLen) and (PS^ in CharSet) and (PS^ <> #0);
    end else
      MatchFail;
end;

constructor TPSECharSet.CreateLen(const SCharSet: string; AMaxLen: integer);
var  j: integer;
     LastChar: Char;
begin
  inherited Create([nil]);
  MaxLen:=AMaxLen;
  if MaxLen < 0 then MaxLen:=0;
  CharSet:=[];
  if Length(SCharSet) > 0 then begin
    j:=1; LastChar:=#0;
    while j <= Length(SCharSet) do begin
      if (SCharSet[j] = '.') and (j in [2..Length(SCharSet)-2])
           and (SCharSet[j+1] = '.') then begin
        CharSet:=CharSet+[LastChar..SCharSet[j+2]];
        LastChar:=SCharSet[j+2];
        Inc(j,3);
      end else begin
        LastChar:=SCharSet[j];
        CharSet:=CharSet+[LastChar];
        Inc(j);
      end;
    end;
  end;
end;

constructor TPSECharSet.Create(const SCharSet: string);
begin
  CreateLen(SCharSet,$7FFFFFFF);
end;

//------------------------------ TPSEOneOf ---------------------------------------

procedure TPSEOneOf.MatchNext;
label L_OK;
var len: integer;
begin
  PS:=PStart;
  while PStack < Pattern.Count do begin
    case Pattern[PStack].ItemType of
      piSimple:
        begin
          len:=Length(Pattern[PStack].pSimple);
          if CompareStr(PChar(Pattern[PStack].pSimple),len) then begin
            PS:=PS+len;
            Inc(PStack);
            goto L_OK;
          end;
        end;
      piPattern:
        begin
          if not CanResume then
            Pattern[PStack].pPattern.ActivatePtr(PS)
          else
            Pattern[PStack].pPattern.MatchNext;
          CanResume:=False;
          if Pattern[PStack].pPattern.IsMatch then begin
            PS:=PS+Pattern[PStack].pPattern.MatchLen;
            goto L_OK;
          end;
        end;
    end;
    Inc(PStack);
  end;
  MatchFail;
  Exit;
L_OK:
  MatchOK;
end;

//-------------------------------- TPSE ----------------------------------------

procedure TPSE.MatchNext;
label L_FAIL;
var len: integer;

  function GoBack: Boolean;
  var j: integer;
  begin
    repeat
      Dec(PStack);
      if PStack < 0 then begin
        CanResume:=False;
        Result:=False;
        Exit;
      end;
      if Pattern[PStack].ItemType = piPattern then
        Pattern[PStack].pPattern.BackValue;
    until (Pattern[PStack].ItemType = piPattern) and
          Pattern[PStack].pPattern.CanResume;
    Result:=(PStack >= 0) and (Pattern[PStack].ItemType = piPattern);
    CanResume:=Result;
    if Result then begin
      PS:=PStart;
      if PStack > 0 then
        for j:=0 to PStack-1 do
          PS:=PS+Pattern[j].MatchLen;
    end;
  end;

begin
  if PStack < 0 then goto L_FAIL;
  if CanResume then
    if not GoBack then goto L_FAIL;
  while PStack < Pattern.Count do begin
    case Pattern[PStack].ItemType of
      piSimple:
        begin
          len:=Length(Pattern[PStack].pSimple);
          if CompareStr(PChar(Pattern[PStack].pSimple),len) then
            PS:=PS+len
          else
            if GoBack then Continue
            else goto L_FAIL;
        end;
      piPattern:
        begin
          if not CanResume then
            Pattern[PStack].pPattern.ActivatePtr(PS)
          else
            Pattern[PStack].pPattern.MatchNext;
          CanResume:=False;
          if Pattern[PStack].pPattern.IsMatch then 
            PS:=PS+Pattern[PStack].pPattern.MatchLen
          else
            if GoBack then Continue
            else goto L_FAIL;
        end;
    end;
    Inc(PStack);
  end;
  MatchOK;
  Exit;
L_FAIL:
  MatchFail;
end;

//------------------------------- TPattern ---------------------------------

function TPattern.GetSetValue(var AValue: string): TPattern;
begin
  Value:=@AValue;
  Result:=Self;
end;

function TPattern.CompareStr(S: PChar; Len: integer): Boolean;
begin
  Result:=StrLComp(S,PS,Len) = 0;
end;

procedure TPattern.MatchFail;
begin
  FMatch:=False;
  CanResume:=False;
  FMatchLen:=0;
  if Value <> nil then Value^:='';
end;

procedure TPattern.MatchOK;
begin
  FMatch:=True;
  CanResume:=True;
  FMatchLen:=PS-PStart;
  if Value <> nil then Value^:=MatchStr;
end;

function TPattern.Match: Boolean;
begin
  while FMatch and (MatchNextPtr^ <> #0) do MatchNext;
  Result:=FMatch and (MatchNextPtr^ = #0);
end;

function TPattern.GetMatchNextPtr: PChar;
begin
  Result:=PStart+MatchLen;
end;

function TPattern.GetMatchStr: string;
begin
  SetLength(Result,MatchLen);
  if MatchLen > 0 then
    StrLCopy(PChar(Result),PStart,MatchLen);
end;

procedure TPattern.Activate(const S: string);
begin
  Line:=S; PStart:=PChar(Line); PS:=PStart;
  Init;
  MatchNext;
end;

procedure TPattern.ActivatePtr(P: PChar);
begin
  Line:=''; PStart:=P; PS:=P;
  Init;
  MatchNext;
end;

constructor TPattern.Create(const APattern: array of const);
begin
  inherited Create;
  Pattern:=TPatList.Create(Self,APattern);
  Value:=nil;
  Line:=''; PStart:=PChar(Line); PS:=PStart;
end;

destructor TPattern.Destroy;
begin
  Pattern.Free;
  inherited Destroy;
end;

procedure TPattern.BackValue;
begin
  if Value <> nil then Value^:=OldValue;
end;

procedure TPattern.Init;
begin
  PStack:=0; FMatch:=False; CanResume:=False;
  if Value <> nil then OldValue:=Value^;
end;

//-------------------------------- TPatList ------------------------------------

constructor TPatList.Create(AOwner: TPattern; const Pattern: array of const);
var j: integer;
begin
  inherited Create;
  Owner:=AOwner;
  for j:=Low(Pattern) to High(Pattern) do
    NewItem(Pattern[j]);
end;

function TPatList.GetItems(Index: integer): TPatItem;
begin
  Result:=TPatItem(inherited Items[Index]);
end;

destructor TPatList.Destroy;
var  Item: TPatItem;
     j: integer;
begin
  for j:=0 to Pred(Count) do begin
    Item:=Items[j];
    Item.Free;
  end;
  inherited Destroy;
end;

procedure TPatList.NewItem(const VarRec: TVarRec);
begin
  inherited Add(TPatItem.Create(VarRec));
end;

//-------------------------------- TPatItem ------------------------------------

constructor TPatItem.Create(const VarRec: TVarRec);
begin
  inherited Create;
  ItemType:=piNil;
  case VarRec.VType of
    vtChar:
      begin
        pSimple:=VarRec.VChar;
        ItemType:=piSimple;
      end;
    vtPChar:
      begin
        pSimple:=VarRec.VPChar;
        ItemType:=piSimple;
      end;
    vtString:
      begin
        pSimple:=VarRec.VString^;
        ItemType:=piSimple;
      end;
    vtAnsiString:
      begin
        pSimple:=string(VarRec.VAnsiString);
        ItemType:=piSimple;
      end;
    vtObject:
      if VarRec.VObject <> nil then
        if VarRec.VObject is TPattern then begin
          pPattern:=TPattern(VarRec.VObject);
          ItemType:=piPattern;
        end;
  end;
end;

function TPatItem.GetMatchLen: integer;
begin
  case ItemType of
    piSimple:  Result:=Length(pSimple);
    piPattern: Result:=pPattern.MatchLen;
    else       Result:=0;
  end;
end;

destructor TPatItem.Destroy;
begin
  if ItemType = piPattern then pPattern.Free;
  inherited Destroy;
end;

end.
